import re
import attr
from   .errors import InvalidFilenameError

# These patterns are interpreted with re.UNICODE in effect, so there's probably
# some character that matches \d but not \w that needs to be included
PYTHON_TAG_RGX   = r'[\w\d]+'
ABI_TAG_RGX      = r'[\w\d]+'
PLATFORM_TAG_RGX = r'[\w\d]+'

WHEEL_FILENAME_CRGX = re.compile(
    r'(?P<project>[A-Za-z0-9](?:[A-Za-z0-9._]*[A-Za-z0-9])?)'
    r'-(?P<version>[A-Za-z0-9_.!+]+)'
    r'(?:-(?P<build>[0-9][\w\d.]*))?'
    r'-(?P<python_tags>{0}(?:\.{0})*)'
    ### XXX: distlib expects only one ABI tag in a filename.  Why?
    r'-(?P<abi_tags>{1}(?:\.{1})*)'
    r'-(?P<platform_tags>{2}(?:\.{2})*)'
    r'\.[Ww][Hh][Ll]'
    .format(PYTHON_TAG_RGX, ABI_TAG_RGX, PLATFORM_TAG_RGX)
)

@attr.s
class ParsedWheelFilename:
    #: The name of the project distributed by the wheel
    project       = attr.ib()
    #: The version of the project distributed by the wheel
    version       = attr.ib()
    #: The wheel's build tag (`None` if not defined)
    build         = attr.ib()
    #: A `list` of Python tags for the wheel
    python_tags   = attr.ib()
    #: A `list` of ABI tags for the wheel
    abi_tags      = attr.ib()
    #: A `list` of platform tags for the wheel
    platform_tags = attr.ib()

    def __str__(self):
        if self.build:
            fmt = '{0.project}-{0.version}-{0.build}-{1}-{2}-{3}.whl'
        else:
            fmt = '{0.project}-{0.version}-{1}-{2}-{3}.whl'
        return fmt.format(
            self,
            '.'.join(self.python_tags),
            '.'.join(self.abi_tags),
            '.'.join(self.platform_tags),
        )

    def tag_triples(self):
        """
        Returns a generator of all simple tag triples formed from the tags in
        the filename
        """
        for py in self.python_tags:
            for abi in self.abi_tags:
                for plat in self.platform_tags:
                    yield '-'.join([py, abi, plat])


def parse_wheel_filename(filename):
    """
    Parse a wheel filename into its components

    :param str filename: a wheel filename
    :rtype: ParsedWheelFilename
    :raises InvalidFilenameError: if the filename is invalid
    """
    m = WHEEL_FILENAME_CRGX.fullmatch(filename)
    if not m:
        raise InvalidFilenameError(filename)
    return ParsedWheelFilename(
        project       = m.group('project'),
        version       = m.group('version'),
        build         = m.group('build'),
        python_tags   = m.group('python_tags').split('.'),
        abi_tags      = m.group('abi_tags').split('.'),
        platform_tags = m.group('platform_tags').split('.'),
    )



import json
from   operator             import attrgetter
from   pathlib              import Path
import pytest
from   wheel_inspect        import Wheel
from   wheel_inspect.errors import WheelValidationError

@pytest.mark.parametrize('whlfile', [
    p for p in (Path(__file__).with_name('data') / 'bad-wheels').iterdir()
      if p.suffix == '.whl'
], ids=attrgetter("name"))
def test_verify_bad_wheels(whlfile):
    with open(str(whlfile.with_suffix('.json'))) as fp:
        expected = json.load(fp)
    with Wheel(str(whlfile)) as whl:
        with pytest.raises(WheelValidationError) as excinfo:
            whl.verify_record()
        assert type(excinfo.value).__name__ == expected["type"]
        assert str(excinfo.value) == expected["str"]



import hashlib
from   packaging.utils import canonicalize_name as normalize

DIGEST_CHUNK_SIZE = 65535

def extract_modules(filelist):
    modules = set()
    for fname in filelist:
        parts = fname.split('/')
        if not parts or not parts[-1].lower().endswith('.py'):
            continue
        parts[-1] = parts[-1][:-3]
        if not all(p.isidentifier() for p in parts):
            continue
        if parts[-1] == '__init__' and len(parts) > 1:
            parts.pop()
        modules.add('.'.join(parts))
    return sorted(modules)

def split_keywords(kwstr):
    # cf. `format_tags()` in Warehouse <https://git.io/fA1AT>, which seems to
    # be the part of PyPI responsible for splitting keywords up for display

    # cf. how wheel handles keywords:
    #keywords = re.split(r'[\0-,]+', kwstr)

    # Based on how pydigger.com seems to handle keywords (See
    # <https://pydigger.com/keywords>):
    if ',' in kwstr:
        return (kwstr.split(','), ',')
        ### TODO: Strip whitespace?  Discard empty keywords?
    else:
        return (kwstr.split(), ' ')

def strfield(s):
    return None if s is None or s.strip() in ('', 'UNKNOWN') else s

def fieldnorm(s):
    return s.lower().replace('-', '_')

def unique_projects(projects):
    seen = set()
    for p in projects:
        pn = normalize(p)
        if pn not in seen:
            yield p
        seen.add(pn)

def digest_file(fp, algorithms):
    digests = {alg: getattr(hashlib, alg)() for alg in algorithms}
    for chunk in iter(lambda: fp.read(DIGEST_CHUNK_SIZE), b''):
        for d in digests.values():
            d.update(chunk)
    return {k: v.hexdigest() for k,v in digests.items()}



import json
from   operator      import attrgetter
from   pathlib       import Path
from   jsonschema    import validate
import pytest
from   wheel_inspect import SCHEMA, inspect_wheel

@pytest.mark.parametrize('whlfile', [
    p for p in (Path(__file__).with_name('data') / 'wheels').iterdir()
      if p.suffix == '.whl'
], ids=attrgetter("name"))
def test_inspect_wheel(whlfile):
    with open(str(whlfile.with_suffix('.json'))) as fp:
        expected = json.load(fp)
    inspection = inspect_wheel(str(whlfile))
    assert inspection == expected
    validate(inspection, SCHEMA)



import pytest
from   wheel_inspect        import ParsedWheelFilename, parse_wheel_filename
from   wheel_inspect.errors import InvalidFilenameError

@pytest.mark.parametrize('filename,parsed', [
    (
        "astrocats-0.3.2-universal-none-any.whl",
        ParsedWheelFilename(
            project='astrocats',
            version='0.3.2',
            build=None,
            python_tags=['universal'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "bencoder.pyx-1.1.2-pp226-pp226-win32.whl",
        ParsedWheelFilename(
            project='bencoder.pyx',
            version='1.1.2',
            build=None,
            python_tags=['pp226'],
            abi_tags=['pp226'],
            platform_tags=['win32'],
        ),
    ),

    (
        "brotlipy-0.1.2-pp27-none-macosx_10_10_x86_64.whl",
        ParsedWheelFilename(
            project='brotlipy',
            version='0.1.2',
            build=None,
            python_tags=['pp27'],
            abi_tags=['none'],
            platform_tags=['macosx_10_10_x86_64'],
        ),
    ),

    (
        "brotlipy-0.3.0-pp226-pp226u-macosx_10_10_x86_64.whl",
        ParsedWheelFilename(
            project='brotlipy',
            version='0.3.0',
            build=None,
            python_tags=['pp226'],
            abi_tags=['pp226u'],
            platform_tags=['macosx_10_10_x86_64'],
        ),
    ),

    (
        "carbonara_archinfo-7.7.9.14.post1-py2-none-any.whl",
        ParsedWheelFilename(
            project='carbonara_archinfo',
            version='7.7.9.14.post1',
            build=None,
            python_tags=['py2'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "coremltools-0.3.0-py2.7-none-any.whl",
        ParsedWheelFilename(
            project='coremltools',
            version='0.3.0',
            build=None,
            python_tags=['py2', '7'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "cvxopt-1.2.0-001-cp34-cp34m-macosx_10_6_intel.macosx_10_9_intel.macosx_10_9_x86_64.macosx_10_10_intel.macosx_10_10_x86_64.whl",
        ParsedWheelFilename(
            project='cvxopt',
            version='1.2.0',
            build='001',
            python_tags=['cp34'],
            abi_tags=['cp34m'],
            platform_tags=[
                'macosx_10_6_intel',
                'macosx_10_9_intel',
                'macosx_10_9_x86_64',
                'macosx_10_10_intel',
                'macosx_10_10_x86_64',
            ],
        ),
    ),

    (
        "django_mbrowse-0.0.1-10-py2-none-any.whl",
        ParsedWheelFilename(
            project='django_mbrowse',
            version='0.0.1',
            build='10',
            python_tags=['py2'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        'efilter-1!1.2-py2-none-any.whl',
        ParsedWheelFilename(
            project='efilter',
            version='1!1.2',
            build=None,
            python_tags=['py2'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "line.sep-0.2.0.dev1-py2.py3-none-any.whl",
        ParsedWheelFilename(
            project='line.sep',
            version='0.2.0.dev1',
            build=None,
            python_tags=['py2', 'py3'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "mayan_edms-1.1.0-1502100955-py2-none-any.whl",
        ParsedWheelFilename(
            project='mayan_edms',
            version='1.1.0',
            build='1502100955',
            python_tags=['py2'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "mxnet_model_server-1.0a5-20180816-py2.py3-none-any.whl",
        ParsedWheelFilename(
            project='mxnet_model_server',
            version='1.0a5',
            build='20180816',
            python_tags=['py2', 'py3'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        'pip-18.0-py2.py3-none-any.whl',
        ParsedWheelFilename(
            project='pip',
            version='18.0',
            build=None,
            python_tags=['py2', 'py3'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "polarTransform-2-1.0.0-py3-none-any.whl",
        ParsedWheelFilename(
            project='polarTransform',
            version='2',
            build='1.0.0',
            python_tags=['py3'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        'psycopg2-2.7.5-cp37-cp37m-macosx_10_6_intel.macosx_10_9_intel.macosx_10_9_x86_64.macosx_10_10_intel.macosx_10_10_x86_64.whl',
        ParsedWheelFilename(
            project='psycopg2',
            version='2.7.5',
            build=None,
            python_tags=['cp37'],
            abi_tags=['cp37m'],
            platform_tags=[
                'macosx_10_6_intel',
                'macosx_10_9_intel',
                'macosx_10_9_x86_64',
                'macosx_10_10_intel',
                'macosx_10_10_x86_64',
            ],
        ),
    ),

    (
        "pyinterval-1.0.0-0-cp27-none-win32.whl",
        ParsedWheelFilename(
            project='pyinterval',
            version='1.0.0',
            build='0',
            python_tags=['cp27'],
            abi_tags=['none'],
            platform_tags=['win32'],
        ),
    ),

    (
        'pypi_simple-0.1.0.dev1-py2.py3-none-any.whl',
        ParsedWheelFilename(
            project='pypi_simple',
            version='0.1.0.dev1',
            build=None,
            python_tags=['py2', 'py3'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "PyQt3D-5.7.1-5.7.1-cp34.cp35.cp36-abi3-macosx_10_6_intel.whl",
        ParsedWheelFilename(
            project='PyQt3D',
            version='5.7.1',
            build='5.7.1',
            python_tags=['cp34', 'cp35', 'cp36'],
            abi_tags=['abi3'],
            platform_tags=['macosx_10_6_intel'],
        ),
    ),

    (
        'qypi-0.4.1-py3-none-any.whl',
        ParsedWheelFilename(
            project='qypi',
            version='0.4.1',
            build=None,
            python_tags=['py3'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "SimpleSteem-1.1.9-3.0-none-any.whl",
        ParsedWheelFilename(
            project='SimpleSteem',
            version='1.1.9',
            build=None,
            python_tags=['3', '0'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "simple_workflow-0.1.47-pypy-none-any.whl",
        ParsedWheelFilename(
            project='simple_workflow',
            version='0.1.47',
            build=None,
            python_tags=['pypy'],
            abi_tags=['none'],
            platform_tags=['any'],
        ),
    ),

    (
        "tables-3.4.2-3-cp27-cp27m-manylinux1_i686.whl",
        ParsedWheelFilename(
            project='tables',
            version='3.4.2',
            build='3',
            python_tags=['cp27'],
            abi_tags=['cp27m'],
            platform_tags=['manylinux1_i686'],
        ),
    ),
])
def test_parse_wheel_filename(filename, parsed):
    assert parse_wheel_filename(filename) == parsed

@pytest.mark.parametrize('filename', [
    "arq-0.3-py35+-none-any.whl",
    "azure_iothub_service_client-1.1.0.0-py2-win32.whl",
    "bgframework-0.4-py2,py3,pypy-none-any.whl",
    "buoyant-0.5.2--py2.py3-none-any.whl",
    "circbuf-0.1b1-py32, py33, py34-none-any.whl",
    "devtools-0.1-py35,py36-none-any.whl",
    "nupic-0.0.31-py2-none-macosx-10.9-intel.whl",
    "qcodes_-0.1.0-py3-none-any.whl",
])
def test_bad_filename(filename):
    with pytest.raises(InvalidFilenameError) as excinfo:
        parse_wheel_filename(filename)
    assert excinfo.value.filename == filename
    assert str(excinfo.value) == 'Invalid wheel filename: ' + repr(filename)



"""
Extract information from wheels

``wheel-inspect`` examines Python wheel files and outputs various information
about the contents within as JSON-serializable objects.  It can be invoked in
Python code as::

    from wheel_inspect import inspect_wheel

    output = inspect_wheel(path_to_wheel_file)

or from the command line with the ``wheel2json`` command.

Visit <https://github.com/jwodder/wheel-inspect> for more information.
"""

from .filename import ParsedWheelFilename, parse_wheel_filename
from .schema   import SCHEMA
from .wheelcls import Wheel, inspect_wheel

__version__      = '1.2.1'
__author__       = 'John Thorvald Wodder II'
__author_email__ = 'wheel-inspect@varonathe.org'
__license__      = 'MIT'
__url__          = 'https://github.com/jwodder/wheel-inspect'

__all__ = [
    'ParsedWheelFilename',
    'SCHEMA',
    'Wheel',
    'inspect_wheel',
    'parse_wheel_filename',
]



from   io                   import StringIO
import json
from   operator             import attrgetter
from   pathlib              import Path
import pytest
from   wheel_inspect.errors import MalformedRecordError
from   wheel_inspect.record import Record

def test_parse_record():
    assert Record.load(StringIO('''\
qypi/__init__.py,sha256=zgE5-Sk8hED4NRmtnPUuvp1FDC4Z6VWCzJOOZwZ2oh8,532
qypi/__main__.py,sha256=GV5UVn3j5z4x-r7YYEB-quNPCucZYK1JOfWxmbdB0N0,7915
qypi/api.py,sha256=2c4EwxDhhHEloeOIeN0YgpIxCGpZaTDNJMYtHlVCcl8,3867
qypi/util.py,sha256=I2mRemqS5PHe5Iabk-CLrgFB2rznR87dVI3YwvpctSQ,3282
qypi-0.4.1.dist-info/DESCRIPTION.rst,sha256=SbT27FgdGvU8QlauLamstt7g4v7Cr2j6jc4RPr7bKNU,11633
qypi-0.4.1.dist-info/LICENSE.txt,sha256=SDaeT4Cm3ZeLgPOOL_f9BliMMHH_GVwqJa6czCztoS0,1090
qypi-0.4.1.dist-info/METADATA,sha256=msK-_0Fe8JHBjBv4HH35wbpUbIlCYv1Vy3X37tIdY5I,12633
qypi-0.4.1.dist-info/RECORD,,
qypi-0.4.1.dist-info/WHEEL,sha256=rNo05PbNqwnXiIHFsYm0m22u4Zm6YJtugFG2THx4w3g,92
qypi-0.4.1.dist-info/entry_points.txt,sha256=t4_O2VB3V-o52_PLoLLIb8m4SQDmY0HFdEJ9_Q2Odtw,45
qypi-0.4.1.dist-info/metadata.json,sha256=KI5TdfaYL-TPS1dMTABV6S8BFq9iAJRk3rkTXjOdgII,1297
qypi-0.4.1.dist-info/top_level.txt,sha256=J2Q5xVa8BtnOTGxjqY2lKQRB22Ydn9JF2PirqDEKE_Y,5
''')).for_json() == [
        {
            "path": "qypi/__init__.py",
            "digests": {
                "sha256": "zgE5-Sk8hED4NRmtnPUuvp1FDC4Z6VWCzJOOZwZ2oh8"
            },
            "size": 532
        },
        {
            "path": "qypi/__main__.py",
            "digests": {
                "sha256": "GV5UVn3j5z4x-r7YYEB-quNPCucZYK1JOfWxmbdB0N0"
            },
            "size": 7915
        },
        {
            "path": "qypi/api.py",
            "digests": {
                "sha256": "2c4EwxDhhHEloeOIeN0YgpIxCGpZaTDNJMYtHlVCcl8"
            },
            "size": 3867
        },
        {
            "path": "qypi/util.py",
            "digests": {
                "sha256": "I2mRemqS5PHe5Iabk-CLrgFB2rznR87dVI3YwvpctSQ"
            },
            "size": 3282
        },
        {
            "path": "qypi-0.4.1.dist-info/DESCRIPTION.rst",
            "digests": {
                "sha256": "SbT27FgdGvU8QlauLamstt7g4v7Cr2j6jc4RPr7bKNU"
            },
            "size": 11633
        },
        {
            "path": "qypi-0.4.1.dist-info/LICENSE.txt",
            "digests": {
                "sha256": "SDaeT4Cm3ZeLgPOOL_f9BliMMHH_GVwqJa6czCztoS0"
            },
            "size": 1090
        },
        {
            "path": "qypi-0.4.1.dist-info/METADATA",
            "digests": {
                "sha256": "msK-_0Fe8JHBjBv4HH35wbpUbIlCYv1Vy3X37tIdY5I"
            },
            "size": 12633
        },
        {
            "path": "qypi-0.4.1.dist-info/RECORD",
            "digests": {},
            "size": None
        },
        {
            "path": "qypi-0.4.1.dist-info/WHEEL",
            "digests": {
                "sha256": "rNo05PbNqwnXiIHFsYm0m22u4Zm6YJtugFG2THx4w3g"
            },
            "size": 92
        },
        {
            "path": "qypi-0.4.1.dist-info/entry_points.txt",
            "digests": {
                "sha256": "t4_O2VB3V-o52_PLoLLIb8m4SQDmY0HFdEJ9_Q2Odtw"
            },
            "size": 45
        },
        {
            "path": "qypi-0.4.1.dist-info/metadata.json",
            "digests": {
                "sha256": "KI5TdfaYL-TPS1dMTABV6S8BFq9iAJRk3rkTXjOdgII"
            },
            "size": 1297
        },
        {
            "path": "qypi-0.4.1.dist-info/top_level.txt",
            "digests": {
                "sha256": "J2Q5xVa8BtnOTGxjqY2lKQRB22Ydn9JF2PirqDEKE_Y"
            },
            "size": 5
        }
    ]

@pytest.mark.parametrize('recfile', [
    p for p in (Path(__file__).with_name('data') / 'bad-records').iterdir()
      if p.suffix == '.csv'
], ids=attrgetter("name"))
def test_parse_bad_records(recfile):
    with open(str(recfile.with_suffix('.json'))) as fp:
        expected = json.load(fp)
    with open(str(recfile), newline='') as fp:
        with pytest.raises(MalformedRecordError) as excinfo:
            Record.load(fp)
        assert type(excinfo.value).__name__ == expected["type"]
        assert str(excinfo.value) == expected["str"]



class WheelValidationError(Exception):
    """ Superclass for all wheel validation errors raised by this package """
    pass


class InvalidFilenameError(WheelValidationError, ValueError):
    """ Raised when an invalid wheel filename is encountered """

    def __init__(self, filename):
        #: The invalid filename
        self.filename = filename

    def __str__(self):
        return 'Invalid wheel filename: ' + repr(self.filename)


class RecordValidationError(WheelValidationError):
    """
    Superclass for all validation errors raised due to a wheel's :file:`RECORD`
    being inaccurate or incomplete
    """
    pass


class RecordSizeMismatchError(RecordValidationError):
    """
    Raised when the size of a file as declared in a wheel's :file:`RECORD` does
    not match the file's actual size
    """

    def __init__(self, path, record_size, actual_size):
        #: The path of the mismatched file
        self.path = path
        #: The size of the file as declared in the :file:`RECORD`
        self.record_size = record_size
        #: The file's actual size
        self.actual_size = actual_size

    def __str__(self):
        return 'Size of file {0.path!r} listed as {0.record_size} in RECORD,'\
               ' actually {0.actual_size}'.format(self)


class RecordDigestMismatchError(RecordValidationError):
    """
    Raised when a file's digest as declared in a wheel's :file:`RECORD` does
    not match the file's actual digest
    """

    def __init__(self, path, algorithm, record_digest, actual_digest):
        #: The path of the mismatched file
        self.path = path
        #: The name of the digest algorithm
        self.algorithm = algorithm
        #: The file's digest as declared in the :file:`RECORD`, in hex
        self.record_digest = record_digest
        #: The file's actual digest, in hex
        self.actual_digest = actual_digest

    def __str__(self):
        return '{0.algorithm} digest of file {0.path!r} listed as'\
               ' {0.record_digest} in RECORD, actually {0.actual_digest}'\
               .format(self)


class FileMissingError(RecordValidationError):
    """
    Raised when a file listed in a wheel's :file:`RECORD` is not found in the
    wheel
    """

    def __init__(self, path):
        #: The path of the missing file
        self.path = path

    def __str__(self):
        return 'File declared in RECORD not found in archive: '+repr(self.path)


class ExtraFileError(RecordValidationError):
    """
    Raised when a wheel contains a file that is not listed in the
    :file:`RECORD` (other than :file:`RECORD.jws` and :file:`RECORD.p7s`)
    """

    def __init__(self, path):
        #: The path of the extra file
        self.path = path

    def __str__(self):
        return 'File not declared in RECORD: {0.path!r}'.format(self)


class MalformedRecordError(WheelValidationError):
    """
    Superclass for all validation errors raised due to a wheel's :file:`RECORD`
    being malformed
    """
    pass


class UnknownDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` uses a digest not listed
    in `hashlib.algorithms_guaranteed`
    """

    def __init__(self, path, algorithm):
        #: The path the entry is for
        self.path = path
        #: The unknown digest algorithm
        self.algorithm = algorithm

    def __str__(self):
        return 'RECORD entry for {0.path!r} uses an unknown digest algorithm:'\
               ' {0.algorithm!r}'.format(self)


class WeakDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` uses a digest weaker than
    sha256
    """

    def __init__(self, path, algorithm):
        #: The path the entry is for
        self.path = path
        #: The weak digest algorithm
        self.algorithm = algorithm

    def __str__(self):
        return 'RECORD entry for {0.path!r} uses a weak digest algorithm:'\
               ' {0.algorithm!r}'.format(self)


class MalformedDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` contains a malformed or
    invalid digest
    """

    def __init__(self, path, algorithm, digest):
        #: The path the entry is for
        self.path = path
        #: The digest's declared algorithm
        self.algorithm = algorithm
        #: The malformed digest
        self.digest = digest

    def __str__(self):
        return 'RECORD contains invalid {0.algorithm} base64 nopad digest for'\
               ' {0.path!r}: {0.digest!r}'.format(self)


class MalformedSizeError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` contains a malformed or
    invalid file size
    """

    def __init__(self, path, size):
        #: The path the entry is for
        self.path = path
        #: The size (as a string)
        self.size = size

    def __str__(self):
        return 'RECORD contains invalid size for {0.path!r}: {0.size!r}'\
               .format(self)


class RecordConflictError(MalformedRecordError):
    """
    Raised when a wheel's :file:`RECORD` contains two or more conflicting
    entries for the same path
    """

    def __init__(self, path):
        #: The path with conflicting entries
        self.path = path

    def __str__(self):
        return 'RECORD contains multiple conflicting entries for {0.path!r}'\
               .format(self)


class EmptyDigestError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has a size but not a
    digest
    """

    def __init__(self, path):
        #: The path the entry is for
        self.path = path

    def __str__(self):
        return 'RECORD entry for {0.path!r} has a size but no digest'\
               .format(self)


class EmptySizeError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has a digest but not a
    size
    """

    def __init__(self, path):
        #: The path the entry is for
        self.path = path

    def __str__(self):
        return 'RECORD entry for {0.path!r} has a digest but no size'\
               .format(self)


class EmptyPathError(MalformedRecordError):
    """ Raised when an entry in a wheel's :file:`RECORD` has an empty path """

    def __str__(self):
        return 'RECORD entry has an empty path'


class RecordLengthError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has the wrong number of
    fields
    """

    def __init__(self, path, length):
        #: The path the entry is for (if nonempty)
        self.path = path
        #: The number of fields in the entry
        self.length = length

    def __str__(self):
        if self.path is None:
            return 'Empty RECORD entry (blank line)'
        else:
            return 'RECORD entry for {0.path!r} has {0.length} fields;'\
                   ' expected 3'.format(self)


class NullEntryError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` lacks both digest and size
    and the entry is not for the :file:`RECORD` itself
    """

    def __init__(self, path):
        #: The path the entry is for
        self.path = path

    def __str__(self):
        return 'RECORD entry for {0.path!r} lacks both digest and size'\
               .format(self)


class NonNormalizedPathError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has a non-normalized path
    """

    def __init__(self, path):
        #: The non-normalized path
        self.path = path

    def __str__(self):
        return 'RECORD entry has a non-normalized path: {0.path!r}'.format(self)


class AbsolutePathError(MalformedRecordError):
    """
    Raised when an entry in a wheel's :file:`RECORD` has an absolute path
    """

    def __init__(self, path):
        #: The absolute path
        self.path = path

    def __str__(self):
        return 'RECORD entry has an absolute path: {0.path!r}'.format(self)


class MissingRecordError(WheelValidationError):
    """ Raised when a wheel does not contain a :file:`RECORD` file """

    def __str__(self):
        return 'No RECORD file in wheel'



import pytest
from   wheel_inspect.util import extract_modules, split_keywords

@pytest.mark.parametrize('kwstr,expected', [
    (
        'pypi,warehouse,search,packages,pip',
        (['pypi', 'warehouse', 'search', 'packages', 'pip'], ','),
    ),
    (
        'pypi warehouse search packages pip',
        (['pypi', 'warehouse', 'search', 'packages', 'pip'], ' '),
    ),
    (
        "pypi,pep503,simple repository api,packages,pip",
        (["pypi", "pep503", "simple repository api", "packages", "pip"], ','),
    ),
])
def test_split_keywords(kwstr, expected):
    assert split_keywords(kwstr) == expected

@pytest.mark.parametrize('filelist,modules', [
    (
        [
            "qypi/__init__.py",
            "qypi/__main__.py",
            "qypi/api.py",
            "qypi/util.py",
            "qypi-0.4.1.dist-info/DESCRIPTION.rst",
            "qypi-0.4.1.dist-info/LICENSE.txt",
            "qypi-0.4.1.dist-info/METADATA",
            "qypi-0.4.1.dist-info/RECORD",
            "qypi-0.4.1.dist-info/WHEEL",
            "qypi-0.4.1.dist-info/entry_points.txt",
            "qypi-0.4.1.dist-info/metadata.json",
            "qypi-0.4.1.dist-info/top_level.txt",
        ],
        ["qypi", "qypi.__main__", "qypi.api", "qypi.util"],
    ),

    (
        [
            "flit/__init__.py",
            "flit/__main__.py",
            "flit/_get_dirs.py",
            "flit/build.py",
            "flit/common.py",
            "flit/inifile.py",
            "flit/init.py",
            "flit/install.py",
            "flit/installfrom.py",
            "flit/log.py",
            "flit/logo.py",
            "flit/sdist.py",
            "flit/upload.py",
            "flit/wheel.py",
            "flit/license_templates/apache",
            "flit/license_templates/gpl3",
            "flit/license_templates/mit",
            "flit/vcs/__init__.py",
            "flit/vcs/git.py",
            "flit/vcs/hg.py",
            "flit/vendorized/__init__.py",
            "flit/vendorized/readme/__init__.py",
            "flit/vendorized/readme/clean.py",
            "flit/vendorized/readme/rst.py",
            "flit-0.11.1.dist-info/entry_points.txt",
            "flit-0.11.1.dist-info/LICENSE",
            "flit-0.11.1.dist-info/WHEEL",
            "flit-0.11.1.dist-info/METADATA",
            "flit-0.11.1.dist-info/RECORD",
        ],
        [
            "flit",
            "flit.__main__",
            "flit._get_dirs",
            "flit.build",
            "flit.common",
            "flit.inifile",
            "flit.init",
            "flit.install",
            "flit.installfrom",
            "flit.log",
            "flit.logo",
            "flit.sdist",
            "flit.upload",
            "flit.vcs",
            "flit.vcs.git",
            "flit.vcs.hg",
            "flit.vendorized",
            "flit.vendorized.readme",
            "flit.vendorized.readme.clean",
            "flit.vendorized.readme.rst",
            "flit.wheel",
        ],
    ),
])
def test_extract_modules(filelist, modules):
    assert extract_modules(filelist) == modules

### TODO: Add more test cases for all functions!



from   cgi                 import parse_header
import io
import os.path
from   zipfile             import ZipFile
from   pkg_resources       import EntryPoint, yield_lines
from   property_manager    import cached_property
from   readme_renderer.rst import render
from   .                   import errors
from   .errors             import WheelValidationError
from   .filename           import parse_wheel_filename
from   .metadata           import parse_metadata
from   .record             import Record
from   .util               import digest_file, extract_modules, \
                                    split_keywords, unique_projects
from   .wheel_info         import parse_wheel_info

def parse_entry_points(fp):
    return {
        gr: {
            k: {
                "module": e.module_name,
                "attr": '.'.join(e.attrs) if e.attrs else None,
                "extras": list(e.extras),
            } for k,e in eps.items()
        } for gr, eps in EntryPoint.parse_map(fp).items()
    }

def readlines(fp):
    return list(yield_lines(fp))

EXTRA_DIST_INFO_FILES = [
    # file name, handler function, result dict key
    # <https://setuptools.readthedocs.io/en/latest/formats.html>:
    ('dependency_links.txt', readlines, 'dependency_links'),
    ('entry_points.txt', parse_entry_points, 'entry_points'),
    ('namespace_packages.txt', readlines, 'namespace_packages'),
    ('top_level.txt', readlines, 'top_level'),
]

class Wheel:
    def __init__(self, path):
        self.path = path
        self.parsed_filename = parse_wheel_filename(os.path.basename(path))
        self.dist_info = '{0.project}-{0.version}.dist-info'\
                            .format(self.parsed_filename)

    def __enter__(self):
        self.fp = open(self.path, 'rb')
        self.zipfile = ZipFile(self.fp)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.zipfile.close()
        self.fp.close()
        return False

    @cached_property
    def record(self):
        rec = self._get_dist_info('RECORD')
        if rec is None:
            raise errors.MissingRecordError()
        with self.zipfile.open(rec) as fp:
            # The csv module requires this file to be opened with `newline=''`
            return Record.load(io.TextIOWrapper(fp, 'utf-8', newline=''))

    def verify_record(self):
        # Check everything in RECORD against actual values:
        for entry in self.record:
            if entry:
                entry.verify(self.zipfile)
            elif entry.path != self.dist_info + '/RECORD':
                raise errors.NullEntryError(entry.path)
        # Check everything in zipfile appears in RECORD (except signatures):
        for path in self.zipfile.namelist():
            if path not in self.record and path not in (
                self.dist_info + '/RECORD.jws',
                self.dist_info + '/RECORD.p7s',
            ):
                raise errors.ExtraFileError(path)

    @cached_property
    def metadata(self):
        rec = self._get_dist_info('METADATA')
        if rec is None:
            ### TODO: This should be an error
            return None
        with self.zipfile.open(rec) as fp:
            return parse_metadata(io.TextIOWrapper(fp, 'utf-8'))

    @cached_property
    def wheel_info(self):
        rec = self._get_dist_info('WHEEL')
        if rec is None:
            ### TODO: This should be an error
            return None
        with self.zipfile.open(rec) as fp:
            return parse_wheel_info(io.TextIOWrapper(fp, 'utf-8'))

    def _get_dist_info(self, filename):
        try:
            return self.zipfile.getinfo(self.dist_info + '/' + filename)
        except KeyError:
            return None

    def inspect(self):
        namebits = self.parsed_filename
        about = {
            "filename": os.path.basename(self.path),
            "project": namebits.project,
            "version": namebits.version,
            "buildver": namebits.build,
            "pyver": namebits.python_tags,
            "abi": namebits.abi_tags,
            "arch": namebits.platform_tags,
        }
        try:
            record = self.record
        except WheelValidationError as e:
            record = None
            about["valid"] = False
            about["validation_error"] = {
                "type": type(e).__name__,
                "str": str(e),
            }
        else:
            try:
                self.verify_record()
            except WheelValidationError as e:
                about["valid"] = False
                about["validation_error"] = {
                    "type": type(e).__name__,
                    "str": str(e),
                }
            else:
                about["valid"] = True

        about["file"] = {"size": os.path.getsize(self.path)}
        self.fp.seek(0)
        about["file"]["digests"] = digest_file(self.fp, ["md5", "sha256"])

        about["dist_info"] = {}
        if self.metadata is not None:
            about["dist_info"]["metadata"] = self.metadata
        if record is not None:
            about["dist_info"]["record"] = record.for_json()
        if self.wheel_info is not None:
            about["dist_info"]["wheel"] = self.wheel_info

        for fname, parser, key in EXTRA_DIST_INFO_FILES:
            info = self._get_dist_info(fname)
            if info is not None:
                with self.zipfile.open(info) as fp:
                    about["dist_info"][key] = parser(io.TextIOWrapper(fp, 'utf-8'))

        if self._get_dist_info('zip-safe') is not None:
            about["dist_info"]["zip_safe"] = True
        elif self._get_dist_info('not-zip-safe') is not None:
            about["dist_info"]["zip_safe"] = False

        md = about["dist_info"].get("metadata", {})
        about["derived"] = {
            "description_in_body": "BODY" in md,
            "description_in_headers": "description" in md,
        }

        if "BODY" in md and "description" not in md:
            md["description"] = md["BODY"]
        md.pop("BODY", None)
        readme = md.get("description")
        if readme is not None:
            md["description"] = {"length": len(md["description"])}
            dct = md.get("description_content_type")
            if dct is None or parse_header(dct)[0] == 'text/x-rst':
                about["derived"]["readme_renders"] = render(readme) is not None
            else:
                about["derived"]["readme_renders"] = True
        else:
            about["derived"]["readme_renders"] = None

        if md.get("keywords") is not None:
            about["derived"]["keywords"], about["derived"]["keyword_separator"] \
                = split_keywords(md["keywords"])
        else:
            about["derived"]["keywords"], about["derived"]["keyword_separator"] \
                = [], None
        about["derived"]["keywords"] = sorted(set(about["derived"]["keywords"]))

        about["derived"]["dependencies"] = sorted(unique_projects(
            req["name"] for req in md.get("requires_dist", [])
        ))

        about["derived"]["modules"] = extract_modules([
            rec["path"] for rec in about["dist_info"].get("record", [])
        ])

        return about

def inspect_wheel(path):
    """
    Examine the Python wheel at the given path and return various information
    about the contents within as a JSON-serializable `dict`
    """
    with Wheel(path) as whl:
        return whl.inspect()



import json
from   operator                 import attrgetter
from   pathlib                  import Path
import pytest
from   wheel_inspect.metadata   import parse_metadata
from   wheel_inspect.wheel_info import parse_wheel_info
from   wheel_inspect.wheelcls   import parse_entry_points

@pytest.mark.parametrize('mdfile', [
    p for p in (Path(__file__).with_name('data') / 'metadata').iterdir()
      if p.suffix == '.metadata'
], ids=attrgetter("name"))
def test_parse_metadata(mdfile):
    with open(str(mdfile.with_suffix('.json'))) as fp:
        expected = json.load(fp)
    with open(str(mdfile), encoding='utf-8') as fp:
        assert parse_metadata(fp) == expected

@pytest.mark.parametrize('epfile', [
    p for p in (Path(__file__).with_name('data') / 'entry_points').iterdir()
      if p.suffix == '.txt'
], ids=attrgetter("name"))
def test_parse_entry_points(epfile):
    with open(str(epfile.with_suffix('.json'))) as fp:
        expected = json.load(fp)
    with open(str(epfile), encoding='utf-8') as fp:
        assert parse_entry_points(fp) == expected

@pytest.mark.parametrize('wifile', [
    p for p in (Path(__file__).with_name('data') / 'wheel_info').iterdir()
      if p.suffix == '.wheel'
], ids=attrgetter("name"))
def test_parse_wheel_info(wifile):
    with open(str(wifile.with_suffix('.json'))) as fp:
        expected = json.load(fp)
    with open(str(wifile), encoding='utf-8') as fp:
        assert parse_wheel_info(fp) == expected
